import warnings
import json
import sys

from unittest2 import TestCase
from mock import patch

import requests
from teleport import *
from werkzeug.exceptions import Unauthorized

from cosmic.exceptions import *
from cosmic.api import API
from cosmic.models import Model, LazyWrapper
from cosmic import api, request

from cosmic import cosmos


cookbook_spec = {
    u'name': u'cookbook',
    u'actions': {
        u'map': {
            u'cabbage': {
                u'accepts': {
                    u'type': u"Struct",
                    u"param": {
                        u"map": {
                            u"spicy": {
                                u"required": True,
                                u"schema": {u"type": u"Boolean"}
                            },
                            u"capitalize": {
                                u"required": False,
                                u"schema": {u"type": u"Boolean"}
                            }
                        },
                        u"order": [u"spicy", u"capitalize"]
                    }
                },
                u'returns': {u'type': u'JSON'},
                u'doc': "Yay cabbage"
            },
            u'noop': {}
        },
        u'order': [u'cabbage', u'noop']
    },
    u"models": [
        {
            u"name": u"Recipe",
            u"schema": {u"type": u"String"}
        },
        {
            u"name": u"Cookie",
            u"schema": {u"type": u"Boolean"}
        }
    ]
}



class TestAPI(TestCase):

    def setUp(self):
        self.maxDiff = None

        self.cookbook = API(u'cookbook')

        @self.cookbook.action(
            accepts=Struct([
                required(u"spicy", Boolean),
                optional(u"capitalize", Boolean)
            ]),
            returns=JSON,
            doc=u"Yay cabbage")
        def cabbage(spicy, capitalize=False):
            if spicy:
                c = "kimchi"
            else:
                c = "sauerkraut"
            if capitalize:
                return c.capitalize()
            else:
                return c

        @self.cookbook.action(accepts=None, returns=None)
        def noop():
            pass

        @self.cookbook.model
        class Recipe(Model):
            schema = String

            @classmethod
            def assemble(cls, datum):
                if datum == "bacon":
                    raise ValidationError("Not kosher")
                return cls(datum)

        @self.cookbook.model
        class Cookie(Model):
            schema = Boolean

        self.app_debug = self.cookbook.get_flask_app(debug=True)
        self.client_debug = self.app_debug.test_client()

        self.app = self.cookbook.get_flask_app()
        self.client = self.app.test_client()

    def test_model(self):
        R = self.cookbook.models.Recipe
        with cosmos:
            self.assertEqual(Schema.to_json(R), {"type": "cookbook.Recipe"})
            pancake = R.from_json("pancake")
            self.assertEqual(pancake.data, u"pancake")
            self.assertEqual(R.to_json(pancake), u"pancake")

    def test_LazyWrapper_okay(self):
        ls = LazyWrapper("cookbook.Recipe")
        with cosmos:
            self.assertEqual(Schema.to_json(ls), {"type": "cookbook.Recipe"})
            # This time model_cls should be cached
            self.assertEqual(Schema.to_json(ls), {"type": "cookbook.Recipe"})

    def test_LazyWrapper_fail(self):
        ls = LazyWrapper("unknown.Unknown")
        with self.assertRaises(ModelNotFound):
            ls.from_json(1)

    # TODO: Teleport should raise a ValidationError here
    def _test_accepts_invalid_schema(self):
        with self.assertRaisesRegexp(ValidationError, "Missing fields"):
            @self.cookbook.action(returns=Schema.from_json({"type": "Struct"}))
            def func(a, b=1):
                pass

    def test_model_deserialize_okay(self):
        with cosmos:
            s = Schema.from_json({"type": "cookbook.Recipe"})
            self.assertEqual(s.from_json("turkey").data, "turkey")

    def test_model_deserialize_bad_name(self):
        with cosmos:
            Schema.from_json({"type": "cookingbook.Recipe"})
            with self.assertRaisesRegexp(ModelNotFound, "cookingbook.Recipe"):
                cosmos.force()

    def test_subclassing_hook(self):
        self.assertEqual(set(self.cookbook.models.__all__), set(["Recipe", "Cookie"]))

    def test_recursive_subclassing_hook(self):
        @self.cookbook.model
        class ChocolateCookie(self.cookbook.models.Cookie):
            pass
        self.assertEqual(set(self.cookbook.models.__all__), set(["Recipe", "Cookie", "ChocolateCookie"]))

    def test_model_schema_validation(self):
        with self.assertRaises(ValidationError):
            self.cookbook.models.Recipe.from_json(1.1)

    def test_model_custom_validation(self):
        with self.assertRaisesRegexp(ValidationError, "kosher"):
            self.cookbook.models.Recipe.from_json("bacon")
        # When not overridden, custom validation passes
        self.cookbook.models.Cookie(True)

    def test_serialize(self):
        self.assertEqual(API.to_json(self.cookbook), cookbook_spec)

    def test_call(self):
        data = '{"spicy": true}'
        self.assertEqual(self.cookbook.actions.cabbage(spicy=False), "sauerkraut")

    def test_spec_endpoint(self):
        res = self.client.get('/spec.json')
        self.assertEqual(json.loads(res.data), cookbook_spec)

    def test_spec_wrong_method(self):
        res = self.client.get('/actions/noop')
        self.assertEqual(res.status_code, 405)
        res = self.client.post('/spec.json')
        self.assertEqual(res.status_code, 405)

    def test_spec_wrong_content_type(self):
        res = self.client.post('/actions/noop')
        self.assertEqual(res.status_code, 400)
        self.assertRegexpMatches(res.data, "Content-Type")
        res = self.client.post('/actions/noop', content_type="application/xml")
        self.assertEqual(res.status_code, 400)
        self.assertRegexpMatches(res.data, "Content-Type")

    def test_schema(self):
        with cosmos:
            API.from_json(API.to_json(self.cookbook))

    def test_load_url(self):
        """Test the API.load function when given a spec URL"""
        with patch.object(requests, 'get') as mock_get:
            mock_get.return_value.text = json.dumps(cookbook_spec)
            mock_get.return_value.status_code = 200
            cookbook_decentralized = API.load('http://example.com/spec.json')
            self.assertEqual(API.to_json(cookbook_decentralized), cookbook_spec)


class TestRemoteAPI(TestCase):

    def setUp(self):
        self.cookbook = API.from_json(cookbook_spec)
        self.cookbook.url = 'http://localhost:8881/api'

    def test_remote_no_return_action(self):
        with patch.object(requests, 'post') as mock_post:
            mock_post.return_value.status_code = 200
            mock_post.return_value.text = ''
            self.assertEqual(self.cookbook.actions.noop(), None)
            mock_post.assert_called_with('http://localhost:8881/api/actions/noop', headers={'Content-Type': 'application/json'}, data="")

    def test_properties(self):
        self.assertEqual(self.cookbook.name, "cookbook")
        self.assertEqual(self.cookbook.url, "http://localhost:8881/api")

    def test_models(self):
        self.assertEqual(self.cookbook.models.__all__, ["Recipe", "Cookie"])
        self.assertEqual(Schema.to_json(self.cookbook.models.Recipe.schema), {"type": "String"})




from werkzeug.local import LocalProxy, LocalStack
from flask import request

from .models import _ctx_stack, _global_cosmos


# Fixed by Armin in cbd8ec9b20405b406a8a701b34087a349b656d5d
# Will go into Werkzeug 0.8.4, remove this hack when it comes out
# http://werkzeug.pocoo.org/docs/changes/#version-0-8-4
class FixedLocalProxy(LocalProxy):
    __coerce__ = lambda x, o: x._get_current_object().__coerce__(x, o) # pragma: no cover
    __enter__ = lambda x: x._get_current_object().__enter__()
    __exit__ = lambda x, *a, **kw: x._get_current_object().__exit__(*a, **kw)

def _get_current_cosmos():
    if _ctx_stack.top != None:
        return _ctx_stack.top
    else:
        return _global_cosmos

cosmos = FixedLocalProxy(_get_current_cosmos)



from unittest2 import TestCase
from mock import patch, Mock

from cosmic.exceptions import *
from cosmic.tools import *
from cosmic import cosmos

from teleport import *


class TestGetArgs(TestCase):

    def test_try_splats(self):
        def f(*arg): pass
        with self.assertRaisesRegexp(SpecError, "splats"):
            get_args(f)

    def test_no_args(self):
        def f(): pass
        self.assertEqual(get_args(f), ((), (),))

    def test_one_arg(self):
        def f(x): pass
        self.assertEqual(get_args(f), (("x",), (),))
        def f(x=1): pass
        self.assertEqual(get_args(f), ((), ("x",),))

    def test_multiple_args(self):
        def f(x, y): pass
        self.assertEqual(get_args(f), (("x", "y",), (),))

    def test_multiple_args_and_kwargs(self):
        def f(x, y=1): pass
        self.assertEqual(get_args(f), (("x",), ("y",),))

    def test_multiple_kwargs(self):
        def f(x=0, y=1): pass
        self.assertEqual(get_args(f), ((), ("x", "y",),))



class TestApplyToFunc(TestCase):

    def test_no_arg_okay(self):
        def f(): return "okay"
        self.assertEqual(apply_to_func(f, None), "okay")

    def test_one_arg_okay(self):
        def f(a): return a
        self.assertEqual(apply_to_func(f, 1), 1)

    def test_one_kwarg_okay(self):
        def f(a=2): return a
        self.assertEqual(apply_to_func(f, 1), 1)
        self.assertEqual(apply_to_func(f, {}), {})
        self.assertEqual(apply_to_func(f, None), 2)

    def test_one_kwarg_passed_none(self):
        # None is an explicit value
        def f(a=2): return a
        n = Box(None)
        self.assertEqual(apply_to_func(f, n), n)

    def test_multiple_args_and_kwargs_okay(self):
        def f(a, b=1): return a, b
        res = apply_to_func(f, {'a': 2})
        self.assertEqual(res, (2, 1,))
        res = apply_to_func(f, {'a': 2, 'b': 2})
        self.assertEqual(res, (2, 2,))

    def test_multiple_kwargs_okay(self):
        def f(a=5, b=1): return a, b
        self.assertEqual(apply_to_func(f, {}), (5, 1,))


class TestPackActionArguments(TestCase):

    def test_one_arg(self):
        res = pack_action_arguments("universe")
        self.assertEqual(res, "universe")

    def test_one_kwarg(self):
        res = pack_action_arguments(what="universe")
        self.assertEqual(res, {"what": "universe"})

    def test_many_kwargs(self):
        res = pack_action_arguments(what="universe", when="now")
        self.assertEqual(res, {"what": "universe", "when": "now"})

    def test_multiple_args(self):
        with self.assertRaises(SpecError):
            pack_action_arguments("universe", "now")

    def test_no_args_no_kwargs(self):
        self.assertEqual(pack_action_arguments(), None)

    def test_mixed_args_and_kwargs(self):
        with self.assertRaises(SpecError):
            pack_action_arguments("universe", when="now")


class TestAssertIsCompatible(TestCase):

    def test_base_cases(self):
        with self.assertRaisesRegexp(SpecError, "needs to accept"):
            assert_is_compatible(JSON, (), ())
        assert_is_compatible(JSON, ("a",), ())
        assert_is_compatible(JSON, (), ("a",))

    def test_not_struct(self):
        with self.assertRaisesRegexp(SpecError, "be a Struct"):
            assert_is_compatible(JSON, ("a", "b",), ())

    def test_missing_required_field(self):
        s = Struct([
            required("a", Integer),
        ])
        with self.assertRaisesRegexp(SpecError, "required field"):
            assert_is_compatible(s, ("a", "b",), ())

    def test_field_should_be_required(self):
        s = Struct([
            required("a", Integer),
            optional("b", Integer)
        ])
        with self.assertRaisesRegexp(SpecError, "required field"):
            assert_is_compatible(s, ("a", "b",), ())

    def test_missing_function_argument(self):
        s = Struct([
            required("a", Integer),
            required("b", Integer),
            optional("c", Integer)
        ])
        with self.assertRaisesRegexp(SpecError, "function argument"):
            assert_is_compatible(s, ("a","b",), ())
        with self.assertRaisesRegexp(SpecError, "function argument"):
            assert_is_compatible(s, ("a",), ("b",))


class TestGetterNamespace(TestCase):

    def setUp(self):
        self.d = OrderedDict([('a', 1,), ('b', 2,), ('c', 3,)])
        self.dispatcher = GetterNamespace(
            get_item=self.d.__getitem__,
            get_all=self.d.keys)

    def test_get_item(self):
        self.assertEqual(self.dispatcher.a, 1)
        self.assertEqual(self.dispatcher.b, 2)
        self.assertEqual(self.dispatcher.c, 3)
        with self.assertRaises(KeyError):
            self.dispatcher.d

    def test_all(self):
        self.assertEqual(self.dispatcher.__all__, ['a', 'b', 'c'])


class TestSchemaHelpers(TestCase):

    def test_deserialize_json(self):
        with self.assertRaisesRegexp(ValidationError, "Expected Box, found None"):
            deserialize_json(Integer, None)
        with self.assertRaisesRegexp(ValidationError, "Expected None, found Box"):
            deserialize_json(None, Box(1))
        self.assertEqual(deserialize_json(None, None), None)
        self.assertEqual(deserialize_json(Integer, Box(1)), 1)

    def test_serialize_json(self):
        with self.assertRaisesRegexp(ValidationError, "Expected data, found None"):
            serialize_json(Integer, None)
        with self.assertRaisesRegexp(ValidationError, "Expected None, found data"):
            serialize_json(None, 1)
        self.assertEqual(serialize_json(None, None), None)
        self.assertEqual(serialize_json(Integer, 1).datum, 1)

    def test_string_to_json(self):
        self.assertEqual(string_to_json(""), None)
        self.assertEqual(string_to_json("1").datum, 1)

    def test_json_to_string(self):
        self.assertEqual(json_to_string(None), "")
        self.assertEqual(json_to_string(Box(1)), "1")



from unittest2 import TestCase

import requests
from mock import patch

from werkzeug.exceptions import Unauthorized, BadRequest, InternalServerError
from flask import Flask

from cosmic.actions import *
from cosmic.tools import *
from cosmic.http import *

from teleport import *


class TestView(TestCase):

    def setUp(self):

        def noop(payload):
            pass
        self.noop = FlaskView(noop, False)

        def unhandled_error(payload):
            return Box(1 / 0)
        self.unhandled_error = FlaskView(unhandled_error, False)

        def authentication_error(payload):
            raise Unauthorized()
        self.authentication_error = FlaskView(authentication_error, False)

        def bad_request(payload):
            raise BadRequest("BOOM")
        self.bad_request = FlaskView(bad_request, False)

        def validation_error(payload):
            raise ValidationError("Invalid!")
        self.validation_error = FlaskView(validation_error, False)

        self.app = Flask(__name__)
        self.app.debug = True

    def test_wrong_content_type(self):
        with self.app.test_request_context('/', method="POST", data="", content_type="application/jason"):
            res = self.noop()
            self.assertEqual(res.status_code, 400)
            self.assertEqual(res.content_type, "application/json")
            self.assertRegexpMatches(res.data, "Content-Type must be")

    def test_invalid_json(self):
        with self.app.test_request_context('/', method="POST", data='{"spicy":farse}', content_type="application/json"):
            res = self.noop()
            self.assertEqual(res.status_code, 400)
            self.assertEqual(res.content_type, "application/json")
            self.assertRegexpMatches(res.data, "Invalid JSON")

    def test_unhandled_error(self):
        with self.app.test_request_context('/', method="POST", data="", content_type="application/json"):
            res = self.unhandled_error()
            self.assertEqual(res.status_code, 500)
            self.assertEqual(res.content_type, "application/json")
            self.assertEqual(json.loads(res.data), {
                "error": "Internal Server Error"
            })

    def test_validation_error(self):
        with self.app.test_request_context('/', method="POST", data="", content_type="application/json"):
            res = self.validation_error()
            self.assertEqual(res.status_code, 400)
            self.assertEqual(res.content_type, "application/json")
            self.assertEqual(json.loads(res.data), {"error": "Invalid!"})

    def test_unhandled_error_debug(self):
        u = FlaskView(
            view=self.unhandled_error.view,
            debug=True)
        with self.assertRaises(ZeroDivisionError):
            with self.app.test_request_context('/', method="POST", data="", content_type="application/json"):
                u()

    def test_HTTPException_handling(self):
        with self.app.test_request_context('/', method="POST", data="", content_type="application/json"):
            res = self.authentication_error()
            self.assertEqual(res.status_code, 401)
            self.assertEqual(res.content_type, "application/json")
            self.assertEqual(json.loads(res.data), {"error": "Unauthorized"})

    def test_HTTPException_with_description_handling(self):
        with self.app.test_request_context('/', method="POST", data="", content_type="application/json"):
            res = self.bad_request()
            self.assertEqual(res.status_code, 400)
            self.assertEqual(res.content_type, "application/json")
            self.assertEqual(json.loads(res.data), {"error": "BOOM"})

    def test_action_no_args_no_data(self):
        with self.app.test_request_context('/', method="POST", data="", content_type="application/json"):
            res = self.noop()
            self.assertEqual(res.status_code, 200)
            self.assertEqual(res.content_type, "application/json")
            self.assertEqual(res.data, "")


class TestCallable(TestCase):

    def setUp(self):
        self.function = Function(Integer, Boolean)
        self.callable = Callable(self.function, "http://example.com/actions/even")

    def test_call_okay(self):
        with patch.object(requests, 'post') as mock_post:
            mock_post.return_value.status_code = 200
            mock_post.return_value.text = "true"
            self.assertEqual(self.callable(1), True)

    def test_call_okay_no_response(self):
        with patch.object(requests, 'post') as mock_post:
            mock_post.return_value.status_code = 200
            mock_post.return_value.text = ""
            self.assertEqual(self.callable(1), None)

    def test_call_server_sent_wrong_type(self):
        with patch.object(requests, 'post') as mock_post:
            mock_post.return_value.status_code = 200
            mock_post.return_value.text = "1"
            with self.assertRaisesRegexp(InternalServerError, "invalid value"):
                self.callable(1)

    def test_call_error_no_message(self):
        with patch.object(requests, 'post') as mock_post:
            mock_post.return_value.status_code = 400
            mock_post.return_value.text = "WTF"
            with self.assertRaises(BadRequest):
                self.callable(1)

    def test_call_error_with_message(self):
        with patch.object(requests, 'post') as mock_post:
            mock_post.return_value.status_code = 500
            mock_post.return_value.text = '{"error": "your mama"}'
            with self.assertRaisesRegexp(InternalServerError, "your mama"):
                self.callable(1)



from __future__ import unicode_literals

import inspect
import json

from .exceptions import SpecError

from teleport import *


class GetterNamespace(object):
    """An object that exposes an arbitrary mapping as a namespace, letting its
    values be accessible as attributes.

    :param get_item: a function that takes a string and returns an item
    :param get_all: a function that returns a list of all keys in the
                    namespace

    Example::

        >>> d = {"a": 1, "b": 2}
        >>> n = GetterNamespace(d.__getitem__, d.keys)
        >>> n.a
        1
        >>> n.__all__
        ["a", "b"]

    """

    def __init__(self, get_item, get_all=None):
        self.get_item = get_item
        self.get_all = get_all

    def __getattr__(self, name):
        return self.get_item(name)

    @property
    def __all__(self):
        return self.get_all()


def get_args(func):
    """Given a function, returns a tuple (*required*, *optional*), tuples of
    non-keyword and keyword arguments respectively. If a function contains
    splats (* or **), a :exc:`~cosmic.exceptions.SpecError` will be raised.
    """
    args, varargs, keywords, defaults = inspect.getargspec(func)
    if varargs or keywords:
        raise SpecError("Cannot define action with splats (* or **)")
    # No arguments
    if len(args) == 0:
        return ((), (),)
    # Number of non-keyword arguments (required ones)
    numargs = len(args) - (len(defaults) if defaults else 0)
    required = tuple(args[:numargs])
    optional = tuple(args[numargs:])
    return (required, optional,)


def apply_to_func(func, data):
    """Applies a piece of normalized data to the user-defined action function
    based on its argument spec. The data is assumed to be normalized by a
    schema compatible with *func*. (see
    :func:`~cosmic.tools.assert_is_compatible`). Thus, no validation is
    performed.

    If *func* takes a single argument, *data* is passed in as is. If it takes
    multiple arguments, *data* is assumed to be a dict and is unpacked into
    the function arguments.

    If object is None, *func* is called with no arguments.
    """
    if data == None:
        return func()
    required, optional = get_args(func)
    # If only one argument, take the whole object
    if len(required + optional) == 1:
        return func(data)
    return func(**data)


def pack_action_arguments(*args, **kwargs):
    """Takes arbitrary args and kwargs and packs them into a dict if there are
    more than one. Returns `None` if there are no arguments. Must be called
    with either a single argument or multiple keyword arguments.
    """
    if len(args) == 1 and len(kwargs) == 0:
        return args[0]
    if len(args) == 0 and len(kwargs) > 0:
        return kwargs
    if len(args) == 0 and len(kwargs) == 0:
        return None
    raise SpecError("Action must be called either with one argument or with one or more keyword arguments")


def assert_is_compatible(schema, required, optional):
    """Raises a :exc:`~cosmic.exceptions.SpecError` if function argument spec
    (as returned by :func:`get_args`) is incompatible with the given schema.
    By incompatible, it is meant that there exists such a piece of data that
    is valid according to the schema, but that could not be applied to the
    function by :func:`apply_to_func`.
    """
    # No arguments
    if len(required + optional) == 0:
        raise SpecError("Function needs to accept arguments")
    # One argument can accept anything
    if len(required + optional) == 1:
        return
    # Multiple arguments means schema must be Struct
    if not isinstance(schema, Struct):
        raise SpecError("For a function that takes arguments, accepts schema"
                        " is expected to be a Struct")
    # Each non-keyword argument in the function must have a corresponding
    # required field in the schema
    for r in required:
        if r not in schema.param.keys() or not schema.param[r]["required"]:
            raise SpecError("Function argument '%s' must have a corresponding"
                            " required field in the accepts schema" % r)
    # All fields in the schema must have a corresponding function argument
    for f in schema.param.keys():
        if f not in set(required + optional):
            raise SpecError("The '%s' field must have a corresponding"
                            " function argument" % f)


def deserialize_json(schema, datum):
    if schema and not datum:
        raise ValidationError("Expected Box, found None")
    if datum and not schema:
        raise ValidationError("Expected None, found Box")
    if schema and datum:
        return schema.from_json(datum.datum)
    return None

def serialize_json(schema, datum):
    if schema and not datum:
        raise ValidationError("Expected data, found None")
    if datum and not schema:
        raise ValidationError("Expected None, found data")
    if schema and datum:
        return Box(schema.to_json(datum))
    return None

def string_to_json(s):
    if s == "":
        return None
    else:
        return Box(json.loads(s))

def json_to_string(box):
    if box == None:
        return ""
    return json.dumps(box.datum)



from __future__ import unicode_literals

import json
import requests

from werkzeug.exceptions import HTTPException, InternalServerError, abort

from flask import Flask, make_response
from flask import request
from teleport import Box, ValidationError

from .tools import *
from .exceptions import *
from . import cosmos


class FlaskView(object):

    def __init__(self, view, debug):
        self.view = view
        self.debug = debug

    def err_to_response(self, err):
        if isinstance(err, HTTPException):
            if err.description != err.__class__.description:
                text = err.description
            else:
                text = err.name
            return self.error_response(text, err.code)
        elif isinstance(err, SpecError):
            return self.error_response(err.args[0], 400)
        elif isinstance(err, ValidationError):
            return self.error_response(str(err), 400)
        else:
            return self.error_response("Internal Server Error", 500)

    def error_response(self, message, code):
        body = json.dumps({"error": message})
        return make_response(body, code, {"Content-Type": "application/json"})

    def __call__(self):
        try:
            ct = request.headers.get('Content-Type', None)
            if request.method != "GET" and ct != "application/json":
                raise SpecError('Content-Type must be "application/json" got %s instead' % ct)
            try:
                request.payload = string_to_json(request.data)
            except ValueError:
                return self.error_response("Invalid JSON", 400)
            with cosmos:
                data = self.view(request.payload)
                body = ""
                if data != None:
                    body = json.dumps(data.datum)
                return make_response(body, 200, {"Content-Type": "application/json"})
        except Exception as err:
            if self.debug:
                raise
            else:
                return self.err_to_response(err)


class Callable(object):

    def __init__(self, function, url):
        self.function = function
        self.url = url

    def __call__(self, *args, **kwargs):
        packed = pack_action_arguments(*args, **kwargs)

        serialized = serialize_json(self.function.accepts, packed)

        data = json_to_string(serialized)

        headers = {'Content-Type': 'application/json'}
        res = requests.post(self.url, data=data, headers=headers)

        try:
            payload = string_to_json(res.text)
        except ValueError:
            payload = None

        if res.status_code != requests.codes.ok:
            message = None
            if payload and type(payload.datum) == dict and 'error' in payload.datum:
                message = payload.datum['error']
            abort(res.status_code, message)
        try:
            if self.function.returns and payload:
                return self.function.returns.from_json(payload.datum)
            else:
                return None
        except ValidationError:
            raise InternalServerError("Call returned an invalid value")



import sys

from werkzeug.local import LocalStack

from teleport import *
from .exceptions import ModelNotFound


class Model(BasicWrapper):
    """A data type definition attached to an API."""

    def __init__(self, data):
        self.data = data

    @classmethod
    def assemble(cls, datum):
        cls.validate(datum)
        return cls(datum)

    @classmethod
    def validate(cls, datum):
        pass

    @classmethod
    def disassemble(cls, datum):
        return datum.data


class LazyWrapper(object):
    _model_cls = None

    def __init__(self, type_name):
        self.type_name = type_name

    def from_json(self, datum):
        return self.model_cls.from_json(datum)

    def to_json(self, datum):
        return self.model_cls.to_json(datum)

    @property
    def model_cls(self):
        from . import cosmos
        if not self._model_cls:
            self._model_cls = cosmos.get_model(self.type_name)
        return self._model_cls


class Cosmos(TypeMap):

    def __init__(self):
        self.lazy_serializers = []
        self.apis = {}

    def force(self):
        for serializer in self.lazy_serializers:
            serializer._model_cls = self.get_model(serializer.type_name)
        self.lazy_serializers = []

    def get_model(self, name):
        api_name, model_name = name.split('.', 1)
        try:
            api = self.apis[api_name]
            model = api._models[model_name]
            return model
        except KeyError:
            raise ModelNotFound(name)

    def __enter__(self):
        _ctx_stack.push(self)
        super(Cosmos, self).__enter__()

    def __exit__(self, *args, **kwargs):
        _ctx_stack.pop()
        super(Cosmos, self).__exit__(*args, **kwargs)

    def __getitem__(self, name):
        from api import API, ModelSerializer
        from actions import Function
        if name == "cosmic.API":
            return (API, None,)
        elif name == "cosmic.Model":
            return (ModelSerializer, None,)
        elif name == "cosmic.Function":
            return (Function, None,)
        elif '.' in name:
            try:
                model_cls = self.get_model(name)
                return (model_cls, None,)
            except ModelNotFound:
                lazy_model = LazyWrapper(name)
                self.lazy_serializers.append(lazy_model)
                return (lazy_model, None,)
        else:
            return BUILTIN_TYPES[name]

_ctx_stack = LocalStack()
# If the stack is empty, use this global object
_global_cosmos = Cosmos()




import random

from cosmic.api import API
from cosmic.models import Model
from teleport import *

zodiac = API("zodiac")

@zodiac.model
class Sign(Model):
    schema = String

    SIGNS = [
        "aries",
        "taurus",
        "gemini",
        "cancer",
        "leo",
        "virgo",
        "libra",
        "scorpio",
        "sagittarius",
        "capricorn",
        "aquarius",
        "pisces"
    ]

    @classmethod
    def validate(cls, datum):
        if datum not in cls.SIGNS:
            raise ValidationError("Unknown zodiac sign", datum)

@zodiac.action(accepts=Sign, returns=String)
def predict(sign):
    ret = "For %s, now is a good time to " % sign.data
    ret += random.choice([
        "build an API.",
        "mow the lawn.",
        "buy Bitcoin.",
        "get a snake tatoo."
    ])
    ret += " It is " + random.choice([
        "probable",
        "not improbable",
        "conceivable",
        "not entirely out of the question"
    ]) + " that you will meet a handsome stranger."
    return ret

if __name__ == "__main__":
    zodiac.run()



from __future__ import unicode_literals

from teleport import *

from .tools import deserialize_json, serialize_json


class Function(BasicWrapper):
    type_name = "cosmic.Function"

    schema = Struct([
        optional("accepts", Schema),
        optional("returns", Schema),
        optional("doc", String)
    ])

    def __init__(self, accepts=None, returns=None, doc=None):
        self.accepts = accepts
        self.returns = returns
        self.doc = doc

    @staticmethod
    def assemble(datum):
        return Function(**datum)

    @staticmethod
    def disassemble(datum):
        return {
            "accepts": datum.accepts,
            "returns": datum.returns,
            "doc": datum.doc
        }

    def json_to_json(self, payload):
        normalized = deserialize_json(self.accepts, payload)
        ret = self.func(normalized)
        return serialize_json(self.returns, ret)



import json
import time
import requests
from mock import patch

from unittest2 import TestCase

from cosmic.api import API
from cosmic import cosmos
from zodiac import zodiac

json_spec = {
    "name": "zodiac",
    "models": [
        {
            "name": "Sign",
            "schema": {"type": "String"}
        }
    ],
    "actions": {
        "map": {
            u"predict": {
                "accepts": {"type": "zodiac.Sign"},
                "returns": {"type": "String"}
            }
        },
        "order": [u"predict"]
    }
}

class TestTutorialBuildingAPI(TestCase):

    def setUp(self):
        self.c = zodiac.get_flask_app().test_client()
        self.d = zodiac.get_flask_app(debug=True).test_client()

    def test_run(self):
        res = self.c.post('/actions/predict', data='"leo"', content_type="application/json")
        self.assertRegexpMatches(res.data, "handsome stranger")

    def test_spec_endpoint(self):
        res = self.c.get('/spec.json')
        self.assertEqual(json.loads(res.data), json_spec)

    def test_wrong_sign(self):
        res = self.c.post('/actions/predict', data='"tiger"', content_type="application/json")
        self.assertEqual(json.loads(res.data), {"error": "Unknown zodiac sign: u'tiger'"})

    def test_consuming(self):
        with patch.object(requests, 'get') as mock_get:
            mock_get.return_value.text = json.dumps(json_spec)
            mock_get.return_value.status_code = 200

            with cosmos:
                h = API.load('http://example.com/spec.json')
                pisces = h.models.Sign("pisces")
                with patch.object(requests, 'post') as mock_post:
                    mock_post.return_value.text = '"Yada yada handsome stranger"'
                    mock_post.return_value.status_code = 200
                    self.assertRegexpMatches(h.actions.predict(pisces), "handsome stranger")


from multiprocessing import Process

def quickstart_test():
    p = Process(target=lambda: zodiac.run(port=9873))
    p.start()
    time.sleep(0.2)
    headers = { 'Content-Type': 'application/json' }
    res = requests.post('http://localhost:9873/actions/predict', data='"leo"', headers=headers)
    assert "handsome" in res.text
    p.terminate()


from unittest2 import TestCase

from teleport import *

from cosmic.models import *
from cosmic.actions import Function
from cosmic.api import API, ModelSerializer 
from cosmic.exceptions import ModelNotFound


class TestCosmos(TestCase):

    def setUp(self):
        self.cosmos = Cosmos()

    def test_cosmic_types(self):
        self.assertEqual(self.cosmos["cosmic.API"], (API, None,))
        self.assertEqual(self.cosmos["cosmic.Model"], (ModelSerializer, None,))
        self.assertEqual(self.cosmos["cosmic.Function"], (Function, None,))

class TestLazyWrapper(TestCase):

    def setUp(self):

        awesome = API("awesome")

        @awesome.model
        class Yay(Model):
            schema = String

        self.Yay = Yay
        self.cosmos = Cosmos()
        self.cosmos.apis["awesome"] = awesome

    def test_from_json(self):

        lazy = LazyWrapper("awesome.Yay")
        self.assertEqual(lazy._model_cls, None)
        with self.cosmos:
            self.assertEqual(lazy.from_json("Bam").data, "Bam")
            self.assertEqual(lazy._model_cls, self.Yay)
            # By now, _model_cls is set
            self.assertEqual(lazy.from_json("Bam").data, "Bam")

    def test_to_json(self):

        lazy = LazyWrapper("awesome.Yay")
        with self.cosmos:
            self.assertEqual(lazy.to_json(self.Yay("Bam")), "Bam")









#!/usr/bin/python

from setuptools import setup

with open("README.rst") as readme:
    long_description = readme.read()

setup(
    name = 'cosmic',
    version = "0.0.6",
    url = "http://www.cosmic-api.com/docs/cosmic/python/0.0/",
    packages = ['cosmic'],
    description = 'A high-level web API framework',
    license = "MIT",
    author = "8313547 Canada Inc.",
    author_email = "alexei.boronine@gmail.com",
    long_description = long_description,
    install_requires = [
        "teleport==0.1.1",
        "Flask==0.9",
        "requests==1.2.3"
    ]
)



from __future__ import unicode_literals

import json
import requests
from multiprocessing import Process
from collections import OrderedDict

from teleport import *

from flask import Blueprint, Flask

from .actions import Function
from .models import Model
from .tools import GetterNamespace, get_args, assert_is_compatible
from .http import FlaskView, Callable
from . import cosmos




class ModelSerializer(BasicWrapper):
    type_name = "cosmic.Model"

    schema = Struct([
        required("name", String),
        optional("schema", Schema)
    ])

    @staticmethod
    def assemble(datum):
        # Take a schema and name and turn them into a model class
        class M(Model):
            schema = datum["schema"]
        M.__name__ = str(datum["name"])
        return M

    @staticmethod
    def disassemble(datum):
        return {
            "name": datum.__name__,
            "schema": datum.schema
        }



class API(BasicWrapper):
    type_name = "cosmic.API"

    schema = Struct([
        required("name", String),
        optional("homepage", String),
        required("models", Array(ModelSerializer)),
        required("actions", OrderedMap(Function))
    ])

    def __init__(self, name, homepage=None, models=[], actions=None):
        self.name = name
        self.homepage = homepage

        if actions:
            self._actions = actions
        else:
            self._actions = OrderedDict()

        self._models = OrderedDict()
        # Populate it if we have initial data
        for model in models:
            model.api = self
            self._models[model.__name__] = model

        self.actions = GetterNamespace(self._get_function_callable)
        self.models = GetterNamespace(
            get_item=self._models.__getitem__,
            get_all=self._models.keys)

        # Add to registry so we can reference its models
        cosmos.apis[self.name] = self

    @staticmethod
    def assemble(datum):
        return API(**datum)

    @staticmethod
    def disassemble(datum):
        return {
            "name": datum.name,
            "homepage": datum.homepage,
            "actions": datum._actions,
            "models": datum._models.values()
        }

    @staticmethod
    def load(url):
        """Given a spec URL, loads the JSON form of an API and deserializes
        it, returning the :class:`~cosmic.api.API` object.
        """
        res = requests.get(url)
        api = API.from_json(json.loads(res.text))
        # Set the API url to be the spec URL, minus the /spec.json
        api.url = url[:-10]
        # Once the API has been added to the cosmos, force lazy models to
        # evaluate.
        cosmos.force()
        return api

    def get_blueprint(self, debug=False):
        """Return a :class:`flask.blueprints.Blueprint` instance containing
        everything necessary to run your API. You may use this to augment an
        existing Flask website with an API::

            from flask import Flask
            from cosmic import API

            hackernews = Flask(__name__)
            hnapi = API("hackernews")

            hackernews.register_blueprint(
                hnapi.get_blueprint(),
                url_prefix="/api")
        
        The *debug* parameter will determine whether Cosmic will propagate
        exceptions, letting them reach the debugger or swallow them up,
        returning proper HTTP error responses.
        """

        def spec_view(payload):
            return Box(self.get_json_spec())

        blueprint = Blueprint('cosmic', __name__)
        blueprint.add_url_rule("/spec.json",
            view_func=FlaskView(spec_view, debug),
            methods=["GET"],
            endpoint="spec")
        for name, function in self._actions.items():
            url = "/actions/%s" % name
            endpoint = "function_%s" % name
            view_func = FlaskView(function.json_to_json, debug)
            blueprint.add_url_rule(url,
                view_func=view_func,
                methods=["POST"],
                endpoint=endpoint)
        return blueprint

    def get_flask_app(self, debug=False, url_prefix=None):
        """Returns a Flask application with nothing but the API blueprint
        registered.
        """
        blueprint = self.get_blueprint(debug=debug)

        app = Flask(__name__, static_folder=None)
        # When debug is True, PROPAGATE_EXCEPTIONS will be implicitly True
        app.debug = debug
        app.register_blueprint(blueprint, url_prefix=url_prefix)

        return app

    def get_json_spec(self):
        return API.to_json(self)


    def run(self, url_prefix=None, api_key=None, registry_url_override=None, **kwargs): # pragma: no cover
        """Runs the API as a Flask app. All keyword arguments except
        *url_prefix* channelled into :meth:`Flask.run`.
        """
        debug = kwargs.get('debug', False)
        app = self.get_flask_app(debug=debug, url_prefix=url_prefix)

        if api_key:
            def register_spec(self, url, api_key, spec):
                import requests
                headers = {'Content-Type': 'application/json'}
                data = {
                    "api_key": api_key,
                    "spec": spec
                }
                requests.post(url, data=data, headers=headers)

            with cosmos:
                spec = API.to_json(self)
            url = "https://registry.cosmic-api.com/actions/register_spec"
            if registry_url_override:
                url = registry_url_override
            p = Process(target=self.register_spec, args=(url, api_key, spec,))
            p.start()
        app.run(**kwargs)


    def action(self, accepts=None, returns=None, doc=None):
        """A decorator for creating actions out of functions and registering
        them with the API.

        The *accepts* parameter is a schema that describes the input of the
        function, *returns* is a schema that describes the output of the
        function. The name of the function becomes the name of the action.

        .. code:: python

            random = API("random")

            @random.action(returns=Integer)
            def generate():
                return 9
        """
        def wrapper(func):
            name = unicode(func.__name__)
            required, optional = get_args(func)

            if accepts:
                assert_is_compatible(accepts, required, optional)

            function = Function(accepts, returns, doc)
            function.func = func

            self._actions[name] = function

            return func
        return wrapper

    def _get_function_callable(self, name):
        function = self._actions[name]
        if hasattr(function, "func"):
            return function.func
        else:
            url = self.url + '/actions/' + name
            return Callable(function, url)

    def model(self, model_cls):
        """A decorator for registering a model with an API. The name of the
        model class is used as the name of the resulting model.

        .. code:: python

            from teleport import String
            from cosmic.models import Model

            dictionary = API("dictionary")

            @dictionary.model
            class Word(Model):
                schema = String

        """
        model_cls.api = self
        model_cls.type_name = "%s.%s" % (self.name, model_cls.__name__,)
        # Add to namespace
        self._models[model_cls.__name__] = model_cls
        return model_cls


